package aql

import (
	"git.science.uu.nl/graphpolaris/query-conversion/entity"
)

// We use consts to define string to prevent typos
const ENTITYSTRING = "entity"
const RELATIONSTRING = "relation"
const GROUPBYSTRING = "groupBy"
const FILTERSTRING = "filter"

var relationdone map[int]bool

func createHierarchy(JSONQuery *entity.IncomingQueryJSON) ([]entity.Tree, entity.QueryEntityStruct) {
	var treeList []entity.Tree
	relationdone = make(map[int]bool)
	topNode := getTopNode(JSONQuery)
	topTreeSelfTriple := getTripleFromNode(JSONQuery, topNode)
	topTree := entity.Tree{
		Self:     topTreeSelfTriple,
		Parent:   -1,
		Children: []int{},
	}
	treeList = append(treeList, topTree)
	treeListIndex := len(treeList) - 1
	treeList = getChildrenFromTree(JSONQuery, treeList, treeListIndex, 0)
	return treeList, topNode
}

/*
Get a entity that has only 1 relation attached an return its index
*/
func getTopNode(JSONQuery *entity.IncomingQueryJSON) entity.QueryEntityStruct {
	indexOfNodeToReturn := -1
	for i, node := range JSONQuery.Entities {
		connectionCount := 0
		for _, relation := range JSONQuery.Relations {
			if (relation.FromType == ENTITYSTRING && relation.FromID == node.ID) || (relation.ToType == ENTITYSTRING && relation.ToID == node.ID) {
				connectionCount++
			}
		}
		if connectionCount == 1 {
			indexOfNodeToReturn = i
			return JSONQuery.Entities[indexOfNodeToReturn]
		}
	}
	return JSONQuery.Entities[indexOfNodeToReturn]
}

func getTripleFromNode(JSONQuery *entity.IncomingQueryJSON, node entity.QueryEntityStruct) entity.Triple {
	var tripleToReturn entity.Triple
	for _, relation := range JSONQuery.Relations {
		// The initial node was our From so we set the Triple accordingly
		// TODO
		// If the To is not an entity we might have to do something different
		if (relation.FromType == ENTITYSTRING && relation.FromID == node.ID) && relation.ToType == ENTITYSTRING {
			tripleToReturn.FromNode = node
			tripleToReturn.Rel = relation
			tripleToReturn.ToNode = JSONQuery.Entities[relation.ToID]
		} else if (relation.ToType == ENTITYSTRING && relation.ToID == node.ID) && relation.FromType == ENTITYSTRING {
			tripleToReturn.FromNode = JSONQuery.Entities[relation.FromID]
			tripleToReturn.Rel = relation
			tripleToReturn.ToNode = node
		}
	}
	relationdone[tripleToReturn.Rel.ID] = true
	return tripleToReturn
}

func getTripleFromRelation(JSONQuery *entity.IncomingQueryJSON, relation entity.QueryRelationStruct) entity.Triple {
	var tripleToReturn entity.Triple
	tripleToReturn.FromNode = JSONQuery.Entities[relation.FromID]
	tripleToReturn.Rel = relation
	tripleToReturn.ToNode = JSONQuery.Entities[relation.ToID]
	return tripleToReturn
}

func getChildrenFromTree(JSONQuery *entity.IncomingQueryJSON, treeList []entity.Tree, treeListIndex int, parentIndex int) []entity.Tree {
	var childRelationTriples []entity.Triple
	for i, relation := range JSONQuery.Relations {
		// We found a relation that is not our parent relation so we can check if it matches on of our nodes
		// If it matches one of the nodes we can add it
		if _, ok := relationdone[i]; !ok {
			if relation.FromType == ENTITYSTRING && relation.FromID == treeList[parentIndex].Self.FromNode.ID {
				triple := getTripleFromRelation(JSONQuery, relation)
				childRelationTriples = append(childRelationTriples, triple)
				relationdone[i] = true
			} else if relation.ToType == ENTITYSTRING && relation.ToID == treeList[parentIndex].Self.ToNode.ID {
				triple := getTripleFromRelation(JSONQuery, relation)
				childRelationTriples = append(childRelationTriples, triple)
				relationdone[i] = true
			}
		}
	}
	// We now have all our children, so we can now make those trees and find their children
	// We can now also add the indices to the list of children from the tree calling this function
	if len(childRelationTriples) != 0 {
		for _, triple := range childRelationTriples {
			childTree := entity.Tree{
				Self:     triple,
				Parent:   parentIndex,
				Children: []int{},
			}
			treeList = append(treeList, childTree)
			// We get the new treeListIndex, which we can now add to the list of children from the tree calling this function
			treeListIndex = len(treeList) - 1
			treeList[parentIndex].Children = append(treeList[parentIndex].Children, treeListIndex)
			treeList = getChildrenFromTree(JSONQuery, treeList, treeListIndex, treeListIndex)
		}
	}
	return treeList
}
